# CSSMinMaxR24.py for Notepad++ Python Scripting plugin * For CSS developers and web designers
# This is a Python port of the YUI CSS Compressor so you can minify/format CSS (JS not tested)
# Based on CSSmin.py v.1.0 from https://github.com/ethanpil/npp-cssmin
#
## Extends 2024
# Patched by Barmaley: bugs fixed + useful features added * GPLv3 * GoodCRX Project
#
# 2024.8.31 (initial fork):
# 1. No required space for @media ") or (" = fixed.
# 2. No required spaces in calc with "+" sign 'calc(1px + 1em)' = fixed.
# 3. Preserved comments now defined with /**..*/ (instead of original /*!..*/)
# 4. If some text selected in active file: process fragment minification in current document (as in original script, see Usage below)
# 5. If nothing selected in active file: force process minification in NEW document (create new than process the copy of source CSS)
# 6. After processing autosave* new compressed file, autonamed from original name with change to "_.css"
#	 * Note: if any file with same name exist in the same dir - it's overwrite without confirmation
# 7. Force pre-save current focument backup copy* (create in same dir as opened file, add .BAK to current filename)
#	 * This copy overwrite every time. No way in NPP to automatically remove temp backup files. Delete them manually if need.
#
# 2024.8.30 -> 2024.8.31:
# + Compressed CSS: all preserved comments delimits by new line from previous/following CSS rules (\n)
# + Fix logic: reset UNDO ops ONLY if minified "by place" (if process run in current file, else undo-buffer stay invalid)
# + Fix/improved: regexp for "calc(x + y)" expressions and for and/or conditions (e.g. @media/@import etc.)
# + Improve/optimize: multi-to-one zeros conversion, multi-semicolon remover, zero-value with units (px/vh/..) fixer
#
# 2024.8.31 -> 2024.9.4:
# + Rebrand project: nows it's named as CSSMinMaxR24
# + Simplest CSS autoformatter: autodetect compressed CSS then expand it - every rules converted to indented CSS blocks:
# 	* If some text selected: replace formatted fragment "by place" (as compressed in previous versions)
# 	* If nothing selected: current compessed CSS expands, create new autonamed file with *_MAX.css as name
#	* Yet works fine only with usual and one-level conditional rules (e.g. @media), otherwise text indents may be not "beauty"
# + Fix/improved: regexp for "calc(x + y)" expressions and for and/or conditions (e.g. @media/@import etc.)
# + Fix/improved: check/autofix for missing/excess :: in pseudo-elements (::after/etc) + fix excess :: in pseudo-classes (:hover/etc)
# + Improve/optimize: multi-to-one zeros conversion, multi-semicolon remover, zero-value with units (px/vh/..) fixer
# + Simplest "CSS-beatifier": add some spaces after/before delimiters (e.g. ,_ :_ _{ ), no trailing spaces and empty lines
# + Source code optimization: some "one-time-call" function moved to aproriate execution points
#
# 2024.9.4 -> 2024.9.7:
# + Comment empty rules in compressor: groups with no declarations, safe "self-backup", no remove em (it will not processed by browser)
# + Fix/improved: check/autofix for missing/excess :: in pseudo-elements and excess ::+ in pseudo-classes (:hover/etc)
# + Post-process usability: after compress/expand - scroll to top if new file created or re-select converted fragment in current file
# + Include wrap-line function to compressor (optional, was unused in source cssmin)
# + Add options block: u can check/change the "set{...}" params to adjust actions (read comments to understand param values).
#	Most of all options adjust the CSS-compressor (to minimize output file size), CSS-expander not required adjustment in general.
#	With changes the option flags you can:
#	* on/off initial backup (force save copy of current document);
#	* control method of empty rules operations (keep/comment/remove);
#	* set your own "important" comments mark (if u want preserv some comments), skip comments in actions (as-is) or force remove em;
#	* note: empty rules and comments can be removed ONLY by compressor, formatter (expander) NOT remove any trash anyway;
#	* control apply experimental features (e.g. pseudo-elements/classes correction);
#	* on/off wrap lines mode and/or beautifier for expander post-processing (improve readability);
#	* set your own mask of compressed/expanded CSS-file (when process entire document and autosave results in a new file);
#	* set every time autosaving before processing, intend only for converts "by place" (if works with selected text)
#	*
#	*
#
## Requirements
#	0. NPP 6.5.1+ (this scrip tested in, older/newer NPP version - test yourself)
#	1. NPP Python Script plugin: install FULL version from https://npppythonscript.sourceforge.net (this script runtime, required)
#	2. NPP Customize Toolbar: install from https://sourceforge.net/projects/npp-customize (optional, best usability if u often use this)
#
## Installation
#	Simply copy "CSSMinMaxR24.py" to "..\plugins\PythonScript\scripts\" in your NPP folder and restart NPP.
#	It will now appear as a menu item under Plugins > Python Script > Scripts
#	For quick access u can adjust NPP Python Script plugin and top level in menu and/or in NPP toolbar
#	For more usability u can also set the icon for this script - e.g. CSSMinMaxR24.bmp (supplied in package)
#	For best usability u can also use* the Customize Toolbar plugin to adjust menu button position.
#		* add the string "Plugins,Python Script,Scripts,CSSMinMaxR24,CSSMinMaxR24.bmp" in CustomizeToolbar.btn file
#
## Usage
#	1. Select text to process, or select nothing.
#	2. Go to NPP menu: Plugins..Python Script..Scripts..CSSMinMaxR24 and click! If u set the script toolbar button - simply click it.
#	3. If some text selected, it will process and replace the selected fragment, otherwise the entire contents of current document.
#	4. Undo is available if u dont like the results.
#	5. Also u can check the output CSS for errors online, e.g. here https://jigsaw.w3.org/css-validator/
#
##Credits
#	Original cssmin.py ported from YUI here https://github.com/zacharyvoase/cssmin
#
##Licence
#	Original BSD changed to GPLv3 for this fork and it's next childs.
###

# Set options: conversion parameters, if not explained - 1 = on, 0 = off.
# Default values tested in practics, don't change em randomly if u not understand the process logic!
set = {
	"bckp": 0,	# Create source CSS file backup before any actions
	# NB! Three flags below are interdepends! Empty rules (if preserved) marks as "important" comments! Be sure u not lost em on changes!
	"ercm": 2,	# Save empty rules as comments. 3: automatic, 2: save all, 1: only with all commented declares, 0: remove all trash
				# Note: if set 3 (auto) - any empty rules or commented declares will be processed by comment rules logic (selective remove)
	"rmcm": 1,	# Skip, process or force remove any comments. 0: skip (stay as-is), 1: automatic with preserv (default), 2: kill em all!
	"impc": "",	# Preserv important (for you) comments from remove in compressor. Default "*" (or any other symbol) = keep that
				# DON'T USE the regexp control symbols! Set any of [^.(){}\[\]!$+/?|] break the comments processing! (and will be skipped)
	"fine": 1,	# Experimental feats: in expander increase readability and remove exseed trash (empty lines, unleased ";", etc.)
	"psel": 1,	# Experimental feats: try to check/correct missing/exceed "::" for pseudo-elements and exceed ":" for pseudo-classes
	"tsas": 0,	# Autosave file every time before run process. Only for "by place" ops (if works with selected text in current file).
	"undo": 0,	# Use new undo buffer (for processing "by place" only). Use 1: ONLY if u sure - avoid reset current document undo buffer!
	"wrap": 0,	# Wrap lines in compressed CSS to an approximate length (increase CSS readability). 0 = no wrap, any integer = max length
				# Try different values to get comfortable text view - e.g. 1: very fine - one line = one rule, 0: 1-line "brick", and so on
	"fmax": "",	# Default file extension prefix mask for autosaved EXPANDED files. If empty - autoset as "_MAX", end ".css" be autoadded
	"fmin": ""	# Default file extension prefix mask for autosaved COMPRESSED files. An mass can be ".min", end ".css" will be autoadded
}

# At first, create source file backup...if u ON this flag... :)
if set["bckp"]:
	notepad.saveAsCopy((notepad.getCurrentFilename() + ".BAK"))

from StringIO import StringIO # The pure-Python StringIO supports unicode.
import re

tmp = ["","","","","","",""]	# Array of temporary strings (used/changed/erased in both minifier/expander)

# Remove CSS comments (except preserved).
def remove_comments(css):
	if re.search(r'[\^\.\(\)\{\}\[\]\*\!\$\+\/\?\|\\]', set["impc"]):	# Skip user-defined comments sign if "preserv" mask is incorrect
		set["impc"] = 0	# Yet set as default "*" = avoid spare dups checks below
	preserve = iemac = 0
	deficm = "*"	# Default mark of preserved comments, also used to preserve empty rules from remove in CSS-compressor
	comment_start = css.find("/*")
	# Slicing to avoid an IndexError.
	while comment_start >= 0:
		# Preserve comments like /**...*/ (by default: two stars ** followed open slash).
		if set["rmcm"] < 2:
			chkpoint = css[comment_start + 2:comment_start + 3]
			preserve = chkpoint == deficm
			if not preserve and set["impc"]:
				preserve = chkpoint == set["impc"]
		comment_end = css.find("*/", comment_start + 2)
		if comment_end < 0:
			if not preserve:
				css = css[:comment_start]
				break
		elif comment_end >= (comment_start + 2):
			if css[comment_end - 1] == "\\":	# Check for IE Mac-specific comment, leave this one and the following one alone.
				comment_start = comment_end + 2
				iemac = 1
			elif iemac:
				comment_start = comment_end + 2
				iemac = 0
			elif not preserve:
				css = css[:comment_start] + css[comment_end + 2:]
			else:
				comment_start = comment_end + 2
		comment_start = css.find("/*", comment_start)
	# Add preserved comments dumb: replaced to \n at last
	tmp[1] = "___PRESCOMM___"
	if set["impc"]:
		css = re.sub(r"(.+)?\/\*"+set["impc"], tmp[1]+"/*"+set["impc"], css)
	return re.sub(r"^"+tmp[1], "", re.sub(r"(.+)?\/\*\*", tmp[1]+"/**", re.sub(r"\*\/\s*", "*/"+tmp[1], css)))

# Remove excess whitespace characters.
def remove_excess_whitespace(css):
	# Prevents 'p :link' from becoming 'p:link': translates 'p :link' into 'p ___PSEUDOCLASSCOLON___link', this is translated back later.
	def pseudoclasscolon(css):
		regex = re.compile(r"(^|\})(([^\{\:])+\:)+([^\{]*\{)")
		match = regex.search(css)
		while match:
			css = ''.join([
				css[:match.start()],
				match.group().replace(":", tmp[2]),
				css[match.end():]])
			match = regex.search(css)
		return css

	# Remove spaces from before things.
	tmp[2] = "___PSEUDOCLASSCOLON___"
	css = pseudoclasscolon(css)
	css = re.sub(r"\s+([!{};:>+\(\)\],])", r"\1", css)

	# If there is a @charset, then only allow one, and move to the beginning.
	css = re.sub(r"^(.*)(@charset \"[^\"]*\";)", r"\2\1", css)
	css = re.sub(r"^(\s*@charset [^;]+;\s*)+", r"\1", css)

	# Put the space back: @media screen and/or (min-width:100px).
	css = re.sub(r"\b(and|or)\(", r"\1 (", css)
	# css = re.sub(r"\bor\(", "or (", css)

	# Put the colons back.
	css = css.replace(tmp[2], ':')

	# Remove spaces from after things.
	css = re.sub(r"([!{}:;>+\(\[,])\s+", r"\1", css)

	# Put the space back: calc(100vw + 100%).
	css = re.sub(r"\b(\d+("+tmp[5]+"))\+", r"\1 + ", css)
	return re.sub(r"\b\)\+", ") + ", css)

# Convert rgb(51,102,153) to #336699.
def normalize_rgb_colors_to_hex(css):
	regex = re.compile(r"rgb\s*\(\s*([0-9,\s]+)\s*\)")
	match = regex.search(css)
	while match:
		colors = map(lambda s: s.strip(), match.group(1).split(","))
		hexcolor = '#%.2x%.2x%.2x' % tuple(map(int, colors))
		css = css.replace(match.group(), hexcolor)
		match = regex.search(css)
	return css

# Replace :0 0 0 0;, :0 0 0; etc. with :0;.
def condense_multidimensional_zeros(css):
	css = re.sub(r"\b:(\s*)?(0\s+){1,3}0;", ":0;", css)
	# Revert background-position:0; to the valid value.
	css = css.replace("background-position:0;", "background-position:0 0;")
	return css

# Shorten colors from #AABBCC to #ABC where possible.
def condense_hex_colors(css):
	regex = re.compile(r"([^\"'=\s])(\s*)#([0-9a-fA-F])([0-9a-fA-F])([0-9a-fA-F])([0-9a-fA-F])([0-9a-fA-F])([0-9a-fA-F])")
	match = regex.search(css)
	while match:
		first = match.group(3) + match.group(5) + match.group(7)
		second = match.group(4) + match.group(6) + match.group(8)
		if first.lower() == second.lower():
			css = css.replace(match.group(), match.group(1) + match.group(2) + '#' + first)
			match = regex.search(css, match.end() - 3)
		else:
			match = regex.search(css, match.end())
	return css

# Wrap the lines of the given CSS to an approximate length.
def wrap_css_lines(css, line_length):
	lines = []
	line_start = 0
	for i, char in enumerate(css):
	# It's safe to break after } characters.
		if char == '}' and (i - line_start >= line_length):
			lines.append(css[line_start:i + 1])
			line_start = i + 1
	if line_start < len(css):
		lines.append(css[line_start:])
	return '\n'.join(lines)

# Try automatic fix for missing/excess ::.
def pseudo_elements_sign(css):
	tmp[4] = "(after|before|marker|selection|placeholder|first-letter|first-line|grammar-error|spelling-error|cue|file-selector-button|highlight|part|slotted|target-text|view-transition)"
	css = re.sub(r"\s?[^:]:("+tmp[4]+"[^\w-])", r"::\1", css)	# Pseudo-elements
	css = re.sub(r"(:\s*)+("+tmp[4]+"[\w-])", r":\2", css)		# Pseudo-classes
	css = re.sub(r"(:\s*)+((?!"+tmp[4]+"))", r":\2", css)		# Pseudo-classes
	css = re.sub(r"(\s*:){2,}\s*", "::", css)	# Pseudo-elements can not be ripped from parent
	return re.sub(r"(:\s*)+((-webkit-|-moz-)[\w-])", r"::\2", css)		# Non-standard pseudo-elements

# CSS-compressor sequence.
def cssmin(css):
	if set["rmcm"]:	# Remove CSS comments (except preserved).
		css = remove_comments(css)
	css = re.sub(r"\s+", " ", css)	# Condense multiple whitespaces into one.
	tmp[3] = "___PSEUDOCLASSBMH___"	# A pseudo class for the Box Model Hack (http://tantek.com/CSS/Examples/boxmodelhack.html)
	css = css.replace('"\\"}\\""', tmp[3])
	tmp[5] = "%|px|vh|vw|vmin|vmax|em|rem|cm|mm|in|pt|pc|ex|ch"
	css = remove_excess_whitespace(css)
	css = re.sub(r"([\s:])(0)("+tmp[5]+"|)", r"\1\2", css)	# Replace 0(px, em, %, etc) with 0.
	css = condense_multidimensional_zeros(css)
	css =  re.sub(r"(:|\s)0+\.(\d+)", r"\1.\2", css)	# Replace 0.6 with .6 where possible.
	css = normalize_rgb_colors_to_hex(css)
	css = condense_hex_colors(css)
	if set["wrap"]:
		css = wrap_css_lines(css, set["wrap"])
	css = css.replace(tmp[3], '"\\"}\\""')
	css = re.sub(r";+\}", "}", re.sub(r";+", ";", css))	# Condense multiple ";;;" into one and remove trailing.
	css = re.sub(tmp[1], r"\n", css)	# Return back the preserved comments with new line indent.
	return css.strip()

# Basic CSS formatter, parse/indents "usual" rules
def base_indent(css):
	tmp[1] = "___AT_COND___"
	tmp[2] = "___AT_RULES_SET___"
	tmp[3] = "___AT_RULES_END___"
	css = re.sub(r'(\*\/)', r'\1\n', css)	# Comments with new lines
	css = re.sub('@', tmp[1], css)	# Preset: conditional rules declare
	css = re.sub(r'('+tmp[1]+'[^\{]+)\{', r'\1'+tmp[2]+'\n', css)	# Preset: conditional rules begin
	css = re.sub('}}', '}'+tmp[3]+'\n', css)	# Preset: conditional rules end
	css = re.sub(r'([^({})]+\{)', r'\1\n\t', css)	# Rules begin
	if set["fine"]:
		css = css_beauty(css, 1)	# Add one spaces after ":"/"," in declares when css-format is basic only
	css = re.sub(r';', r';\n\t', css)		# Next declare
	css = re.sub(r'(?!"\'[^;]*);(?![^;]*\'")', r';\n\t', css)		# Next declare
		# css = re.sub(r'(\t*\})', r'\n\1\n', css)	# Rules end
	css = re.sub(r'(\t*\})', r'\n}\n', css)	# Rules end
	if set["fine"]:
		css = css_beauty(css, 2)	# Add one space after ":" in declares and absent ";" at last declaration end
	return css

# Conditional rules formatter (parse/indents @media/etc. attached rules blocks)
def cond_indent(css):
	def rule_shift(css, arr, cond, shift):	# Note: shift yet works fine ONLY for ONE-level conditional rules!
		shift = '\t'*shift
		for rule in arr:	# Check for rules block is inside the condition
			if re.search(rgx, rule):	# Add more tab-shift to all internal rules and declares strings
				css = css.replace(rule, rule.replace('\n', '\n'+shift))
			else:	# Add tab-shift before all conditional rules and declares strings
				css = css.replace(rule, rule+'\n\t')
		return css

	rgx = r'('+tmp[3]+'|'+tmp[2]+')'
	css = rule_shift(css, re.split(tmp[1], css), tmp[3], 1)	# Note: now shift works fine ONLY for ONE-level conditional rules!
	# Revert back temp conditional rules presets to source
	css = re.sub(r'(\t)?'+tmp[1], '@', css)	# Revert declares preset
	css = re.sub(tmp[2], ' {', css)			# Revert begins preset
	css = re.sub(r'(\t)?'+tmp[3], '}', css)		# Revert ends preset
	return css

# CSS beautifier: increase readability of expanded text.
def css_beauty(css, step):
	if step == 1:	# Pre-improve when css-format is basic only (avoid unwanted issues at next process steps)
		css = re.sub(r'(["\']data:)(.+;)(base64,)', r'\1 \2\n\t\t\3 ', css)	# Experimental: add space after ":"/"," in base64-urls
	if step == 2:	# Pre-improve before run conditional formatter, after all basic formats applied (avoid issues at next steps)
		css = re.sub(r'(\n\t[^:]+:)', r'\1 ', css)	# Add absent space after ":" in declares
		css = re.sub(r'([^\};]*)(?!\*+\/)\n\}', r'\1;\n}', css)	# Add absent ";" at last declaration end
	else:	# Final improvments before autosave
		css = re.sub('{', ' {', css)	# Add spaces before "{" for best readability
		css = re.sub(r'(?!\'data:\s?image\/[a-z]{3,4};base64),', ', ', css)	# Add spaces after "," for best readability
		css = re.sub(r'(\n\s*;*)+\n', r'\n', css)	# Remove exceed ";" and empty lines
		css = re.sub(r'(\s+)$', '', re.sub(r'[ ]{2,}', ' ', css))	# Remove exceed spaces and trailing spaces/tabs
		css = css+"\n"	# Add line at the EOF
	return css

# CSS expander/formatter: expand compressed CSS to human-view format.
def cssmax(css):
	css = base_indent(css)		# Basic formatter, "usual" rules
	css = cond_indent(css)		# Conditional rules formatter
	if set["fine"]:
		css = css_beauty(css,0)	# Improve readability
	return css

# Remove or preserv empty rules (safe "self-backup", no remove).
def empty_rules(css):
	r1 = r2 = ""
	if set["ercm"] > 1:
		r1 = r"\1/*this-is-empty:rule;*/}"
	if set["ercm"] > 0:
		r2 = r"/**\1\2\3*/"
	css = re.sub(r"([^}{]+{)([\s;]*)(})", r1, css)	# Empty rules with no declares (has selector only)
	css = re.sub(r"([^}{]+{[\s]*)[/*]+([^/*]*)[*/]+([\s]*})", r2, css)	# Rules with all commented declares
	return css

# Preprocess by flag max|min.
def outcss(css, asf):
	if asf:				# Autosave current file before run process, every time (if set and process selected text).
		notepad.save()	# If process all text in a new file - force autosave source file, anyway.
	if set["ercm"] < 3:	# Process empty rules
		css = empty_rules(css)
	max = re.search("\}((\n[^\/])|([^\/]\n))", css)	# Set process direction flag (max|min) then run conversion.
	if max:	# Set global temp-strings array for CSS-minifier
		if set["fmin"]:
			tmp[0] = tmp[0]+set["fmin"]
		css = cssmin(css)
	else:	# Set global temp-strings array for CSS-expander
		if set["fmax"]:
			tmp[0] = tmp[0]+set["fmax"]
		else:
			tmp[0] = tmp[0]+"MAX"
		css = cssmax(css)
	if set["psel"]:	# Fix pseudo-elements rules (optional)
		css = pseudo_elements_sign(css)
	return css

## All functions defined. Now execute.
csstxt = editor.getSelText()	# Is selected text - process that, otherwise - process entire document

if len(csstxt):	# If selection in current document: process fragment "by place"
	tmp[6] = editor.getSelectionStart()	# Usability: mem the start selection before conversion
	if set["undo"]:	# New undo buffer: reset current document undo buffer!
		editor.beginUndoAction()
	editor.replaceSel(outcss(csstxt, set["tsas"]))
	if set["undo"]:
		editor.endUndoAction()
	# Usability: re-select converted fragment after conversion
	tmp[5] = editor.getCurrentPos()
	editor.setSelectionStart(tmp[6])
	editor.setSelectionEnd(tmp[5])
else:	# If nothing selected: set the new filename to create copy as new document
	tmp[6] = outcss(editor.getText(), 1)
	tmp[5] = notepad.getCurrentFilename()
	notepad.new()
	editor.setText(tmp[6])
	notepad.saveAs(re.sub(r"\.css$", "_"+tmp[0]+".css", tmp[5]))
	editor.gotoLine(0)	# Usability: scroll to top after conversion

tmp = ["_","","","","","",""]	# Reset/cleanup temp-strings array

# That's all, folks! :)